
from optparse import OptionParser
import time
import sys

import uspp.uspp as serial
import intelhex

__version__ = 0.1


class FBoot:
	REVISION	= 0		# get bootloader revision
	BUFFSIZE	= 1 	# get buffer size
	SIGNATURE	= 2 	# get target signature
	USERFLASH	= 3		# get user flash size
	PROGRAM		= 4		# program flash
	START		= 5		# start application
	CHECK_CRC	= 6		# CRC o.k.
	VERIFY		= 7		# Verify
	
	
	COMMAND		= 0xA5		# Command sequence start
	ESCAPE		= COMMAND
	
	CONNECT		= 0xA6		# connection established
	BADCOMMAND	= 0xA7		# command not supported
	ANSWER		= 0xA8		# followed by length byte
	CONTINUE	= 0xA9
	SUCCESS		= 0xAA
	FAIL		= 0xAB
	
	ESC_SHIFT	= 0x80		# offset escape char
	PROGEND		= ESC_SHIFT
	
	
	TIMEOUT		= 0.3
	TIMEOUTP	= 4
	STARTWAIT	= 3
	MAXFLASH	= 0x40000
	
	def __init__(self, port, baudrate, password = "Peda", debug = False):
		
		self.port = port
		self.baudrate = baudrate
		
		self.crc = 0
		self.echocount = 0
		self.debug = debug
		
		self.buffersize = 0
		self.flashsize = 0
		
		# open serial port
		print "%s at %i Baud" % (self.port, self.baudrate)
		
		interface = serial.SerialPort(self.port, timeout = 100, speed = self.baudrate)
		interface.flush()
		
		self.interface = interface
		
		# try to connect to the bootloader
		echo = False
		
		password += chr(0xff)
		pos = 0
		
		waitstring = ['|', '/', '-', '\\']
		wait = 0
		while True:
			self.interface.write(password[pos])
			pos += 1
			if pos >= len(password):
				pos = 0
			
			sys.stdout.write("\b%s" % waitstring[wait])
			sys.stdout.flush()
			
			wait += 1
			if wait >= len(waitstring):
				wait = 0
			
			i = self.receive()
			if i == password[0]:
				echo = True				# echo received
			
			if i == self.CONNECT:
				if echo:
					self.echocount = 1
					print "\bConnected (One wire)"
				else:
					self.echocount = 0
					print "\bConnected"
				
				# clear the rx buffer
				self.send_command( self.COMMAND )
				while True:
					result = self.receive()
					if result == -1 or result == self.SUCCESS:
						self.interface.flush()
						break
				break
			
			time.sleep(0.1)
	
	def readval(self):
		""" read a number
		
		format:
		ANSWER, length, ... databyte * (length - 1) ..., SUCCESS """
		
		missing = 0xff
		val = 0
		
		while True:
			data = self.receive(10)
			if data < 0:
				return False
			
			if missing == 0xff:
				# nothing received right now
				if data == self.FAIL:
					return -2
				elif data == self.ANSWER:
					missing = 0xfe
			
			elif missing == 0xfe:
				# save number of characters to receive
				missing = data
				if missing > 4:
					# something is going wrong
					return -2
			
			elif missing > 1:
				val = (val << 8) + data
				missing -= 1
			
			else:
				# last character
				if data == self.SUCCESS:
					return val
				else:
					return False
	
	def read_info(self):
		""" Get some information about the bootloader """
		
		self.send_command(self.REVISION)
		i = self.readval()
		
		print "Bootloader V%x.%x" % (i>>8, i & 0xff)
		
		self.send_command(self.SIGNATURE)
		i = self.readval()
		
		targets = {
			0x1e9007: "ATtiny13",
			0x1e910A: "ATtiny2313",
			0x1e9206: "ATtiny45",
			0x1e9205: "ATmega48",
			0x1e9307: "ATmega8",
			0x1e9403: "ATmega16",
			0x1e9406: "ATmega168",
			0x1e9502: "ATmega32",
			0x1e9609: "ATmega644",
			0x1e9802: "ATmega2561",
		}
		print "Target: %06x %s" % (i, targets[i])

		self.send_command(self.BUFFSIZE)
		i = self.readval()
		
		if i == -1:
			return 1
		
		self.buffersize = i
		print "Buffer: %d Byte" % i

		self.send_command(self.USERFLASH)
		i = self.readval()
		
		if i == -1 or i > self.MAXFLASH:
			return 1
		
		self.flashsize = i
		print "Size available: %d Byte" % i
	
	def start_app(self):
		self.send_command(self.PROGEND)
		self.send_command(self.COMMAND)
		self.send_command(self.START)
	
	def check_crc(self, output = False):
		self.send_command(self.CHECK_CRC)
		if self.receive() == self.BADCOMMAND:
			return 0
		
		self.send(self.crc & 0xff)
		self.send(self.crc >> 8)
		
		i = self.receive()
		if i == self.SUCCESS:
			if output: print "CRC: o.k."
			return 2
		else:
			if output: print "CRC: error !"
			return 1
	
	def send(self, c):
		if self.echocount:
			if self.echount > 1:
				self.receive()
			self.echocount += 1
		if self.debug: print "-> 0x%02x" % c
		
		self.interface.write(chr(c))
		
		# calculate new crc-value
		crc = self.crc ^ c
		for j in xrange(8):
			if crc & 0x1:
				crc = (crc >> 1) ^ 0xA001
			else:
				crc = crc >> 1
		self.crc = crc & 0xffff
	
	def send_command(self, c):
		if self.echocount:
			self.echocount = 1
		
		self.send(self.COMMAND)
		self.send(c)
	
	def receive(self, retry = 1):
		while True:
			try:
				data = self.interface.read()
				if len(data) > 0:
					if self.debug: print "<- 0x%02x" % ord(data)
					if self.echocount <= 1:
						return ord(data)
					self.echocount -= 1
			
			except serial.SerialPortException, e:
				retry -= 1
				if retry == 0:
					return -1
	
	def program(self, filename, verify = False):
		
		# read hex-file
		hexfile = intelhex.IntelHexParser(filename)
		if len(hexfile.segments) > 1:
			raise Exception("Currently not more than Segment allowed!")
		hex_data = [ord(x) for x in hexfile.segments[0].data]
		
		# check if there are to much data for the avr
		if len(hex_data) > self.flashsize:
			raise Exception("Hex-file to large for target!")
		
		if not verify:
			print "Program Size: %i Byte" % len(hex_data)
			print "Program %s: 00000 - 00000" % filename,
			self.send_command(self.PROGRAM)
		else:
			# check if we could perform a verify operation
			self.send_command(self.VERIFY)
			if self.receive() == self.BADCOMMAND:
				print "Verify not available"
				return 0
			
			print "Verify  %s: 00000 - 00000" % filename,
		sys.stdout.flush()
		
		# send the hole program
		addr = 0
		bufferpos = self.buffersize
		while True:
			# get one byte
			byte = hex_data[addr]
			
			# escape and send it
			if byte == self.ESCAPE or byte == 0x13:
				self.send(self.ESCAPE)
				byte |= self.ESC_SHIFT
			self.send(byte)
			
			bufferpos -= 1
			if bufferpos == 0:
				print "\b\b\b\b\b\b%05x" % (addr + 1),
				sys.stdout.flush()
				
				if not verify and self.receive(10) != self.CONTINUE:
					print " failed!"
					return 1
				
				bufferpos = self.buffersize
			
			addr += 1
			if addr == len(hex_data):
				print "\b\b\b\b\b\b%05x" % addr,
				
				# send the end-record (0xA5, 0x80)
				self.send_command(self.PROGEND)
				
				if self.receive(10) == self.SUCCESS:
					print " successful"
					return 0
				else:
					print " failed!"
					return 1

if __name__ == '__main__':
	parser = OptionParser(
			usage   = "%prog [options] FILE",
			version = "%prog version: " + str(__version__) )
	parser.add_option("-p", "--port", dest="port",
			default="/dev/ttyUSB0",
			help="serial port (default is '/dev/ttyUSB0')")
	parser.add_option("-b", "--baud", dest="baudrate", 
			type="int",
			default="38400",
			help="baudrate (default is '38400')")
	parser.add_option("-v", "--verify", dest="verify",
			action="store_true",
			help="Perform Verify")
	parser.add_option("-n", "--no-program", dest="only_verify",
			action="store_true",
			help="Don't write the program")
	parser.add_option("-d", "--debug", dest="debug",
			action="store_true",
			help="print some debugging information")
	
	options, args = parser.parse_args()
	
	if len(args) != 1:
		parser.error("incorrect number of arguments")
	
	try:
		# connect to the bootloader
		bootloader = FBoot(options.port, options.baudrate, debug = options.debug)
		
		crc_on = bootloader.check_crc()
		bootloader.read_info()
		
		#if crc_on and bootloader.check_crc(True) == 2:
		#	raise Exception("CRC-Error")
		
		# save current time
		starttime = time.time()
		
		if not options.only_verify:
			bootloader.program(args[0])
		
		if options.verify:
			bootloader.program(args[0], verify = True)
		
		#if crc_on:
		#	bootloader.check_crc(True)
		#else:
		#	print "CRC: not available"
		
		bootloader.start_app()
		
		print "Elapsed time: %4.2f seconds" % (time.time() - starttime)
	except KeyboardInterrupt:
		print " Aborted!"
	#except Exception, e:
	#	print e
